package s_com.udojava.evalex;

import java.util.Iterator;
import java.util.Stack;
import java.util.ArrayList;
import java.util.Objects;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.math.MathContext;
import java.math.BigDecimal;

public class Expression
{
    public static final BigDecimal PI;
    private MathContext mc;
    private String expression;
    private List<String> rpn;
    private Map<String, Operator> operators;
    private Map<String, Function> functions;
    private Map<String, BigDecimal> variables;
    private final char decimalSeparator = '.';
    private final char minusSign = '-';
    
    public Expression(final String expression) {
        this.mc = MathContext.DECIMAL32;
        this.expression = null;
        this.rpn = null;
        this.operators = new HashMap<String, Operator>();
        this.functions = new HashMap<String, Function>();
        this.variables = new HashMap<String, BigDecimal>();
        this.expression = expression;
        this.addOperator(new Operator("+", 20, true) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return v1.add(v2, Expression.this.mc);
            }
        });
        this.addOperator(new Operator("-", 20, true) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return v1.subtract(v2, Expression.this.mc);
            }
        });
        this.addOperator(new Operator("*", 30, true) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return v1.multiply(v2, Expression.this.mc);
            }
        });
        this.addOperator(new Operator("/", 30, true) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return v1.divide(v2, Expression.this.mc);
            }
        });
        this.addOperator(new Operator("%", 30, true) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return v1.remainder(v2, Expression.this.mc);
            }
        });
        this.addOperator(new Operator("^", 40, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, BigDecimal v2) {
                final int signOf2 = v2.signum();
                final double dn1 = v1.doubleValue();
                v2 = v2.multiply(new BigDecimal(signOf2));
                final BigDecimal remainderOf2 = v2.remainder(BigDecimal.ONE);
                final BigDecimal n2IntPart = v2.subtract(remainderOf2);
                final BigDecimal intPow = v1.pow(n2IntPart.intValueExact(), Expression.this.mc);
                final BigDecimal doublePow = new BigDecimal(Math.pow(dn1, remainderOf2.doubleValue()));
                BigDecimal result = intPow.multiply(doublePow, Expression.this.mc);
                if (signOf2 == -1) {
                    result = BigDecimal.ONE.divide(result, Expression.this.mc.getPrecision(), RoundingMode.HALF_UP);
                }
                return result;
            }
        });
        this.addOperator(new Operator("&&", 4, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                final boolean b1 = !v1.equals(BigDecimal.ZERO);
                final boolean b2 = !v2.equals(BigDecimal.ZERO);
                return (b1 && b2) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator("||", 2, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                final boolean b1 = !v1.equals(BigDecimal.ZERO);
                final boolean b2 = !v2.equals(BigDecimal.ZERO);
                return (b1 || b2) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator(">", 10, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return (v1.compareTo(v2) == 1) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator(">=", 10, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return (v1.compareTo(v2) >= 0) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator("<", 10, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return (v1.compareTo(v2) == -1) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator("<=", 10, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return (v1.compareTo(v2) <= 0) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator("=", 7, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return (v1.compareTo(v2) == 0) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator("==", 7, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return Expression.this.operators.get("=").eval(v1, v2);
            }
        });
        this.addOperator(new Operator("!=", 7, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return (v1.compareTo(v2) != 0) ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addOperator(new Operator("<>", 7, false) {
            @Override
            public BigDecimal eval(final BigDecimal v1, final BigDecimal v2) {
                return Expression.this.operators.get("!=").eval(v1, v2);
            }
        });
        this.addFunction(new Function("NOT", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final boolean zero = parameters.get(0).compareTo(BigDecimal.ZERO) == 0;
                return zero ? BigDecimal.ONE : BigDecimal.ZERO;
            }
        });
        this.addFunction(new Function("IF", 3) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final boolean isTrue = !parameters.get(0).equals(BigDecimal.ZERO);
                return isTrue ? parameters.get(1) : parameters.get(2);
            }
        });
        this.addFunction(new Function("RANDOM", 0) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.random();
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("SIN", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.sin(Math.toRadians(parameters.get(0).doubleValue()));
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("COS", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.cos(Math.toRadians(parameters.get(0).doubleValue()));
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("TAN", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.tan(Math.toRadians(parameters.get(0).doubleValue()));
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("SINH", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.sinh(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("COSH", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.cosh(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("TANH", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.tanh(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("RAD", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.toRadians(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("DEG", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.toDegrees(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("MAX", 2) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final BigDecimal v1 = parameters.get(0);
                final BigDecimal v2 = parameters.get(1);
                return (v1.compareTo(v2) > 0) ? v1 : v2;
            }
        });
        this.addFunction(new Function("MIN", 2) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final BigDecimal v1 = parameters.get(0);
                final BigDecimal v2 = parameters.get(1);
                return (v1.compareTo(v2) < 0) ? v1 : v2;
            }
        });
        this.addFunction(new Function("ABS", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                return parameters.get(0).abs(Expression.this.mc);
            }
        });
        this.addFunction(new Function("LOG", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.log(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("LOG10", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final double d = Math.log10(parameters.get(0).doubleValue());
                return new BigDecimal(d, Expression.this.mc);
            }
        });
        this.addFunction(new Function("ROUND", 2) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final BigDecimal toRound = parameters.get(0);
                final int precision = parameters.get(1).intValue();
                return toRound.setScale(precision, Expression.this.mc.getRoundingMode());
            }
        });
        this.addFunction(new Function("FLOOR", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final BigDecimal toRound = parameters.get(0);
                return toRound.setScale(0, RoundingMode.FLOOR);
            }
        });
        this.addFunction(new Function("CEILING", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final BigDecimal toRound = parameters.get(0);
                return toRound.setScale(0, RoundingMode.CEILING);
            }
        });
        this.addFunction(new Function("SQRT", 1) {
            @Override
            public BigDecimal eval(final List<BigDecimal> parameters) {
                final BigDecimal x = parameters.get(0);
                if (x.compareTo(BigDecimal.ZERO) == 0) {
                    return new BigDecimal(0);
                }
                if (x.signum() < 0) {
                    throw new ExpressionException("Argument to SQRT() function must not be negative");
                }
                final BigInteger n = x.movePointRight(Expression.this.mc.getPrecision() << 1).toBigInteger();
                final int bits = n.bitLength() + 1 >> 1;
                BigInteger ix = n.shiftRight(bits);
                BigInteger ixPrev;
                do {
                    ixPrev = ix;
                    ix = ix.add(n.divide(ix)).shiftRight(1);
                    Thread.yield();
                } while (ix.compareTo(ixPrev) != 0);
                return new BigDecimal(ix, Expression.this.mc.getPrecision());
            }
        });
        this.variables.put("PI", Expression.PI);
        this.variables.put("TRUE", BigDecimal.ONE);
        this.variables.put("FALSE", BigDecimal.ZERO);
    }
    
    private boolean isNumber(final String st) {
        final char char1 = st.charAt(0);
        Objects.requireNonNull(this);
        if (char1 == '-' && st.length() == 1) {
            return false;
        }
        for (final char ch : st.toCharArray()) {
            if (!Character.isDigit(ch)) {
                final char c = ch;
                Objects.requireNonNull(this);
                if (c != '-') {
                    final char c2 = ch;
                    Objects.requireNonNull(this);
                    if (c2 != '.') {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    
    private List<String> shuntingYard(final String expression) {
        final List<String> outputQueue = new ArrayList<String>();
        final Stack<String> stack = new Stack<String>();
        final Tokenizer tokenizer = new Tokenizer(expression);
        String lastFunction = null;
        String previousToken = null;
        while (tokenizer.hasNext()) {
            final String token = tokenizer.next();
            if (this.isNumber(token)) {
                outputQueue.add(token);
            }
            else if (this.variables.containsKey(token)) {
                outputQueue.add(token);
            }
            else if (this.functions.containsKey(token.toUpperCase())) {
                stack.push(token);
                lastFunction = token;
            }
            else if (Character.isLetter(token.charAt(0))) {
                stack.push(token);
            }
            else if (",".equals(token)) {
                while (!stack.isEmpty() && !"(".equals(stack.peek())) {
                    outputQueue.add(stack.pop());
                }
                if (stack.isEmpty()) {
                    throw new ExpressionException("Parse error for function '" + lastFunction);
                }
            }
            else if (this.operators.containsKey(token)) {
                final Operator o1 = this.operators.get(token);
                for (String token2 = stack.isEmpty() ? null : stack.peek(); this.operators.containsKey(token2) && ((o1.isLeftAssoc() && o1.getPrecedence() <= this.operators.get(token2).getPrecedence()) || o1.getPrecedence() < this.operators.get(token2).getPrecedence()); token2 = (stack.isEmpty() ? null : stack.peek())) {
                    outputQueue.add(stack.pop());
                }
                stack.push(token);
            }
            else if ("(".equals(token)) {
                if (previousToken != null && this.isNumber(previousToken)) {
                    throw new ExpressionException("Missing operator at character position " + tokenizer.getPos());
                }
                stack.push(token);
            }
            else if (")".equals(token)) {
                while (!stack.isEmpty() && !"(".equals(stack.peek())) {
                    outputQueue.add(stack.pop());
                }
                if (stack.isEmpty()) {
                    throw new RuntimeException("Mismatched parentheses");
                }
                stack.pop();
                if (!stack.isEmpty() && this.functions.containsKey(stack.peek().toUpperCase())) {
                    outputQueue.add(stack.pop());
                }
            }
            previousToken = token;
        }
        while (!stack.isEmpty()) {
            final String element = stack.pop();
            if ("(".equals(element) || ")".equals(element)) {
                throw new RuntimeException("Mismatched parentheses");
            }
            if (!this.operators.containsKey(element)) {
                throw new RuntimeException("Unknown operator or function: " + element);
            }
            outputQueue.add(element);
        }
        return outputQueue;
    }
    
    public BigDecimal eval() {
        final Stack<BigDecimal> stack = new Stack<BigDecimal>();
        for (final String token : this.getRPN()) {
            if (this.operators.containsKey(token)) {
                final BigDecimal v1 = stack.pop();
                final BigDecimal v2 = stack.pop();
                stack.push(this.operators.get(token).eval(v2, v1));
            }
            else if (this.variables.containsKey(token)) {
                stack.push(this.variables.get(token).round(this.mc));
            }
            else if (this.functions.containsKey(token.toUpperCase())) {
                final Function f = this.functions.get(token.toUpperCase());
                final ArrayList<BigDecimal> p = new ArrayList<BigDecimal>(f.getNumParams());
                for (int i = 0; i < f.numParams; ++i) {
                    p.add(0, stack.pop());
                }
                final BigDecimal fResult = f.eval(p);
                stack.push(fResult);
            }
            else {
                stack.push(new BigDecimal(token, this.mc));
            }
        }
        return stack.pop().stripTrailingZeros();
    }
    
    public Expression setPrecision(final int precision) {
        this.mc = new MathContext(precision);
        return this;
    }
    
    public Expression setRoundingMode(final RoundingMode roundingMode) {
        this.mc = new MathContext(this.mc.getPrecision(), roundingMode);
        return this;
    }
    
    public Operator addOperator(final Operator operator) {
        return this.operators.put(operator.getOper(), operator);
    }
    
    public Function addFunction(final Function function) {
        return this.functions.put(function.getName(), function);
    }
    
    public Expression setVariable(final String variable, final BigDecimal value) {
        this.variables.put(variable, value);
        return this;
    }
    
    public Expression setVariable(final String variable, final String value) {
        if (this.isNumber(value)) {
            this.variables.put(variable, new BigDecimal(value));
        }
        else {
            this.expression = this.expression.replaceAll("\\b" + variable + "\\b", "(" + value);
            this.rpn = null;
        }
        return this;
    }
    
    public Expression with(final String variable, final BigDecimal value) {
        return this.setVariable(variable, value);
    }
    
    public Expression and(final String variable, final String value) {
        return this.setVariable(variable, value);
    }
    
    public Expression and(final String variable, final BigDecimal value) {
        return this.setVariable(variable, value);
    }
    
    public Expression with(final String variable, final String value) {
        return this.setVariable(variable, value);
    }
    
    public Iterator<String> getExpressionTokenizer() {
        return new Tokenizer(this.expression);
    }
    
    private List<String> getRPN() {
        if (this.rpn == null) {
            this.rpn = this.shuntingYard(this.expression);
        }
        return this.rpn;
    }
    
    public String toRPN() {
        String result = new String();
        for (String st : this.getRPN()) {
            result = (result.isEmpty() ? result : result);
            result += st;
        }
        return result;
    }
    
    static {
        PI = new BigDecimal("3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679");
    }
    
    public class ExpressionException extends RuntimeException
    {
        private static final long serialVersionUID = 1118142866870779047L;
        
        public ExpressionException(final String message) {
            super(message);
        }
    }
    
    public abstract class Function
    {
        private String name;
        private int numParams;
        
        public Function(final String name, final int numParams) {
            this.name = name.toUpperCase();
            this.numParams = numParams;
        }
        
        public String getName() {
            return this.name;
        }
        
        public int getNumParams() {
            return this.numParams;
        }
        
        public abstract BigDecimal eval(final List<BigDecimal> p0);
    }
    
    public abstract class Operator
    {
        private String oper;
        private int precedence;
        private boolean leftAssoc;
        
        public Operator(final String oper, final int precedence, final boolean leftAssoc) {
            this.oper = oper;
            this.precedence = precedence;
            this.leftAssoc = leftAssoc;
        }
        
        public String getOper() {
            return this.oper;
        }
        
        public int getPrecedence() {
            return this.precedence;
        }
        
        public boolean isLeftAssoc() {
            return this.leftAssoc;
        }
        
        public abstract BigDecimal eval(final BigDecimal p0, final BigDecimal p1);
    }
    
    private class Tokenizer implements Iterator<String>
    {
        private int pos;
        private String input;
        private String previousToken;
        
        public Tokenizer(final String input) {
            this.pos = 0;
            this.input = input.trim();
        }
        
        @Override
        public boolean hasNext() {
            return this.pos < this.input.length();
        }
        
        private char peekNextChar() {
            if (this.pos < this.input.length() - 1) {
                return this.input.charAt(this.pos + 1);
            }
            return '\0';
        }
        
        @Override
        public String next() {
            final StringBuilder token = new StringBuilder();
            if (this.pos >= this.input.length()) {
                return this.previousToken = null;
            }
            char ch;
            for (ch = this.input.charAt(this.pos); Character.isWhitespace(ch) && this.pos < this.input.length(); ch = this.input.charAt(++this.pos)) {}
            if (Character.isDigit(ch)) {
                while (true) {
                    if (!Character.isDigit(ch)) {
                        final char c = ch;
                        Objects.requireNonNull(Expression.this);
                        if (c != '.') {
                            break;
                        }
                    }
                    if (this.pos >= this.input.length()) {
                        break;
                    }
                    token.append(this.input.charAt(this.pos++));
                    ch = ((this.pos == this.input.length()) ? '\0' : this.input.charAt(this.pos));
                }
            }
            else {
                final char c2 = ch;
                Objects.requireNonNull(Expression.this);
                if (c2 == '-' && Character.isDigit(this.peekNextChar()) && ("(".equals(this.previousToken) || ",".equals(this.previousToken) || this.previousToken == null || Expression.this.operators.containsKey(this.previousToken))) {
                    final StringBuilder sb = token;
                    Objects.requireNonNull(Expression.this);
                    sb.append('-');
                    ++this.pos;
                    token.append(this.next());
                }
                else if (Character.isLetter(ch) || ch == '_') {
                    while ((Character.isLetter(ch) || Character.isDigit(ch) || ch == '_') && this.pos < this.input.length()) {
                        token.append(this.input.charAt(this.pos++));
                        ch = ((this.pos == this.input.length()) ? '\0' : this.input.charAt(this.pos));
                    }
                }
                else if (ch == '(' || ch == ')' || ch == ',') {
                    token.append(ch);
                    ++this.pos;
                }
                else {
                    while (!Character.isLetter(ch) && !Character.isDigit(ch) && ch != '_' && !Character.isWhitespace(ch) && ch != '(' && ch != ')' && ch != ',' && this.pos < this.input.length()) {
                        token.append(this.input.charAt(this.pos));
                        ++this.pos;
                        final char c3;
                        ch = (c3 = ((this.pos == this.input.length()) ? '\0' : this.input.charAt(this.pos)));
                        Objects.requireNonNull(Expression.this);
                        if (c3 == '-') {
                            break;
                        }
                    }
                    if (!Expression.this.operators.containsKey(token.toString())) {
                        throw new ExpressionException("Unknown operator '" + token + "' at position " + (this.pos - token.length() + 1));
                    }
                }
            }
            return this.previousToken = token.toString();
        }
        
        @Override
        public void remove() {
            throw new ExpressionException("remove() not supported");
        }
        
        public int getPos() {
            return this.pos;
        }
    }
}
